unit cDIB;

{-----------------------------------------------------------------------------
The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/MPL-1.1.html

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either expressed or implied. See the License for
the specific language governing rights and limitations under the License.

The Original Code is: cDIB.PAS, released August 28, 2000.

The Initial Developer of the Original Code is Peter Morris (pete@stuckindoors.com),
Portions created by Peter Morris are Copyright (C) 2000 Peter Morris.
All Rights Reserved.

Purpose of file:
This is the main drawing engine.

Contributor(s):
None as yet


Last Modified: September 01, 2000
Current Version: 1.00

You may retrieve the latest version of this file at the Project JEDI home page,
located at http://www.delphi-jedi.org
or at http://www.stuckindoors.com/dib


Known Issues:
Need to be able to add image compression to lower the size of applications.
Once implemented, reading should check if the data starts with the const
cCompressedSIG, if so read compressed, otherwise read uncompressed.  All writing
from that point on should write compressed (lossless !!!)
-----------------------------------------------------------------------------}


{$O-} //This is needed as some routines are called implicitly, and will be omitted
      //by the compiler
{$A+} //This is default anyway, but we need alignment for DIBS, and it wont hurt.

interface

uses
  Classes, Windows, SysUtils, Graphics, Math, JPeg, Dialogs, cDIBPalette, forms;

type
  EDIBError = Class(Exception);

  TAngle = 0..359;
  TDIBFilter = class;

  //A blitter proc is a routine which copies data from 1 DIB to another.
  //New blitter procs may be written.  To activate the blitter proc you will
  //need to override ChangeBlitter.
  TBlitterProc = procedure(SourceData, DestData         : Pointer;
                           SourceModulo, DestModulo     : DWord;
                           NoPixels, NoLines            : Integer) of object;

  //The main class that all DIBs are created from, this holds all functions for
  //manipulating the data, but does not actually create any data
  TAbstractSuperDIB = class(TPersistent)
  private
    FAngle                    : TAngle;
    FAutoSize                 : Boolean;
    FClipRect                 : TRect;
    FHeight                   : Word;
    FMasked                   : Boolean;
    FOpacity                  : Byte;
    FOwnedData                : Boolean;
    FScaleX,
    FScaleY                   : Real;
    FTransparent              : Boolean;
    FTransparentColor         : TColor;
    FTransparentMode          : TTransparentMode;
    FWidth                    : Word;
    FOnChange                 : TNotifyEvent;

    procedure FreeTheData;
    procedure CreateTheData;
    //Can't see why you would want to override these property routines !
    procedure SetClipRect(const aRect : TRect);
    procedure SetTransparent(const Value: Boolean);
    procedure SetTransparentColor(const Value : TColor);
    procedure SetTransparentMode(const Value : TTransparentMode);
    function GetTransparentColor: TColor;
  protected
    //Which blitter routine to use for drawing
    FBlitter                  : TBlitterProc;
    FData                     : Pointer;

    //Abstract method which MUST be overridden
    procedure CreateData; virtual; abstract;
    procedure FreeData; virtual; abstract;

    //Streaming
    procedure DefineProperties(Filer : TFiler); override;

    //OnChange notification, and an opportunity to alter the blitter proc
    procedure Changed; virtual;
    procedure ChangeBlitter; virtual;

    //Blitter routines
    //Only blit the mask
    procedure BlitMaskOnly    (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    //Copy the bytes, no special effects
    procedure SolidBlit       (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    procedure SolidBlitO      (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    //Copy the bytes, take into account the Mask value
    procedure MaskedBlit      (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    procedure MaskedBlitO     (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    //Blit the image, but not Transparent Colors
    procedure TransparentBlit (SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;
    procedure TransparentBlitO(SourceData, DestData         : Pointer;
                               SourceModulo, DestModulo     : DWord;
                               NoPixels, NoLines            : Integer); virtual;

    //Load the data from a stream
    procedure LoadDataFromStream(S : TStream); virtual;
    //Save the data to a stream
    procedure SaveDataToStream(S : TStream); virtual;


    //property routines
    procedure SetAngle(const Value : TAngle); virtual;
    procedure SetAutoSize(const Value : Boolean); virtual;
    procedure SetHeight(const aValue : Word); virtual;
    procedure SetMasked(const Value : Boolean); virtual;
    procedure SetOpacity(const Value: Byte); virtual;
    procedure SetScale(const Value: Real); virtual;
    procedure SetScaleX(const Value: Real); virtual;
    procedure SetScaleY(const Value: Real); virtual;
    procedure SetWidth(const aValue : Word); virtual;

    property Angle            : TAngle
      read FAngle
      write SetAngle;
    property AutoSize         : Boolean
      read FAutoSize
      write SetAutoSize;
    property ClipRect         : TRect
      read FClipRect
      write SetClipRect;
    property Height           : Word
      read FHeight
      write SetHeight;
    property Masked           : Boolean
      read FMasked
      write SetMasked;
    property Opacity          : Byte
      read FOpacity
      write SetOpacity;
    property Scale            : Real
      read FScaleX
      write SetScale;
    property ScaleX           : Real
      read FScaleX
      write SetScaleX;
    property ScaleY           : Real
      read FScaleY
      write SetScaleY;
    property Transparent      : Boolean
      read FTransparent
      write SetTransparent;
    property TransparentColor : TColor
      read GetTransparentColor
      write SetTransparentColor;
    property TransparentMode  : TTransparentMode
      read FTransparentMode
      write SetTransparentMode;
    property Width            : Word
      read FWidth
      write SetWidth;

    property OnChange         : TNotifyEvent
      read FOnChange
      write FOnChange;
  public
    //Standard create routines
    constructor Create; overload; Virtual;
    constructor Create(aWidth, aHeight : Word); overload; Virtual;
    //This constructor creates a new header for manipulation, but points FData
    //at existing data from another DIB.  This allows 2 DIBs to Share data.
    //There will be a problem if the source DIB is freed !
    constructor CreateReplicaOf(aSource : TAbstractSuperDIB); virtual;
    destructor Destroy; override;

    //Apply a 3x3 matrix
    procedure ApplyFilter(AFilter : TDIBFilter); virtual;
    //Like AssignTo, but ONLY for header information, not data
    procedure AssignHeaderTo(Dest : TPersistent); virtual;
    procedure AssignTo(Dest : TPersistent); override;
    //Like AssignTo, but ONLY the data, not the header
    procedure CopyPicture(Source : TAbstractSuperDIB);
    procedure StretchCopyPicture(Source : TAbstractSuperDIB);

    //Draw will calculate the parameters for the BlitterProc, and then call
    //the current Blitter routine
    procedure Draw(DestX, DestY           : Integer;
                   DestWidth, DestHeight  : Integer;
                   Dest                   : TAbstractSuperDIB;
                   SrcX, SrcY             : Word); virtual;
    //This will copy the mask only, not RGB values, it uses BlitMaskOnly
    procedure DrawMask(DestX, DestY           : Integer;
                       DestWidth, DestHeight  : Integer;
                       Dest                   : TAbstractSuperDIB;
                       SrcX, SrcY             : Word); virtual;

    //Import a mask from a file.  The file MUST be 8bit GreyScale
    procedure ImportMask(aFilename : String);
    //Import a picture from any supported file
    procedure ImportPicture(aFilename : String);
    //Converts the DIB to a RGN by using the MASK
    function MakeRGN(const aMasklevel : Byte) : HRGN;
    //Converts the DIB to a RGN by using a TColor
    function MakeRGNFromColor(aColor : TColor) : HRGN;
    //Destroys the current data, and references an existing DIB's data
    procedure PointDataAt(aSource : TAbstractSuperDIB); virtual;
    //Fill the DIB with a color
    procedure QuickFill(aColor : TColor); virtual;
    //Fill a sub-rect of the DIB with a color
    procedure QuickFillRect(aColor : TColor;  aLeft, aTop, aWidth, aHeight : Integer); virtual;
    //Render8BIT will render to an 8BIT dc using a DIBPalette with UseTable=True
    procedure Render8Bit(DestDC: HDC; X, Y, aWidth,
                         aHeight: Integer; XSrc, YSrc: Word; ROP: Cardinal; Palette: TDIBPalette); virtual;
    //Override ResetHeader in child classes if you add new properties
    procedure ResetHeader; virtual;
    procedure ReSize(aWidth, aHeight : Word); virtual;
    //This routine handles Rotate AND Zoom at the same time, Good eh ?
    procedure RotoZoom(D : TAbstractSuperDIB); virtual;
    //Set all the masked values in the DIB to Opacity
    procedure SetMaskedValues(Const Opacity : Byte);
    //Is the DIB drawable or not
    function  Valid : Boolean; virtual;
  published
  end;

  TCustomWinDIB = class(TAbstractSuperDIB)
  private
    FDC                       : HDC;
    FBitmap                   : HBitmap;
    FOldBitmap                : HBitmap;
    FCanvas                   : TCanvas;
  protected
    procedure CreateData; override;
    procedure FreeData; override;

    property Canvas           : TCanvas
      read FCanvas;
    property Handle           : HDC
      read FDC;
  public
    constructor Create; override;
    destructor Destroy; override;
  published
  end;

  TWinDIB = class(TCustomWinDIB)
  private
  protected
  public
    property ClipRect;
    property Data : Pointer read FData;
    property ScaleX;
    property ScaleY;
  published
    property Angle;
    property AutoSize;
    property Canvas;
    property Handle;
    property Height;
    property Masked;
    property Opacity;
    property Scale;
    property Transparent;
    property TransparentColor;
    property TransparentMode;
    property Width;
  end;

  //A MemoryDIB uses the GlobalHeap to store its data.
  //(No handles for example)
  TCustomMemoryDIB = class(TAbstractSuperDIB)
  private
  protected
    procedure CreateData; override;
    procedure FreeData; override;
  public
  published
  end;

  TMemoryDIB = class(TCustomMemoryDIB)
  private
  protected
  public
    property ClipRect;
    property ScaleX;
    property ScaleY;
  published
    property Angle;
    property AutoSize;
    property Height;
    property Masked;
    property Opacity;
    property Scale;
    property Transparent;
    property TransparentColor;
    property TransparentMode;
    property Width;
  end;

  //A DIBFilter will apply a 3x3 matrix to a DIB.  Simple !
  TDIBFilter = class(TPersistent)
  private
    FRedBias,
    FGreenBias,
    FBlueBias                 : Smallint;
    FOpacity                  : Byte;
    FFactor                   : Integer;
  protected
  public
    Data                      : Array[0..8] of SmallInt;
    constructor Create; virtual;
    class function GetDisplayName : String;

  published
    property BlueBias         : Smallint
      read FBlueBias
      write FBlueBias;
    property GreenBias        : Smallint
      read FGreenBias
      write FGreenBias;
    property Factor           : Integer
      read FFactor
      write FFactor;
    property Opacity          : Byte
      read FOpacity
      write FOpacity;
    property RedBias          : Smallint
      read FRedBias
      write FRedBias;
  end;



  function GetRotatedSize(Width, Height : Word; Angle : TAngle; ScaleX, ScaleY : Real) : TPoint;
  function RelativeAngle(X1,Y1, X2,Y2 : Integer) : Integer;


var
  SinTable1,
  CosTable1,
  SinTable2,
  CosTable2       : Array[0..High(TAngle)] of integer;

implementation

const
  cCompressedSIG = 'DIBCMP';
  cRectAllocs = 400;
  cMaxRectChunks = 2000;

type
  PByte = ^Byte;
  PDWord = ^DWord;

function LastError: string;
var
  OutputMessage: PChar;
begin
  FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM or FORMAT_MESSAGE_ALLOCATE_BUFFER,
    nil,
    GetLastError,
    0,
    @OutputMessage,
    0,
    nil);

  Result := string(OutputMessage);
end;



function GetRotatedSize(Width, Height : Word; Angle : TAngle; ScaleX, ScaleY : Real) : TPoint;
var
  Radians                     : Double;
  ScaleW,
  ScaleH                      : Double;
begin
  Radians := DegToRad(-Angle);
  ScaleW := Width * ScaleX / 100;
  ScaleH := Height * ScaleY / 100;
//  Result.X := Abs(Round(Abs(Width * Cos(Radians)) - Abs(Height * Sin(Radians))));
//  Result.Y := Abs(Round(Abs(Width * Sin(Radians)) + Abs(Height * Cos(Radians))));
  Result := Point(Round(Abs(ScaleW * Cos(Radians)) + Abs(ScaleH * Sin(Radians))),
                  Round(Abs(ScaleW * Sin(Radians)) + Abs(ScaleH * Cos(Radians))));
end;
(*
var
  ScaleHeight,
  ScaleWidth,
  Theta             : Real;
  RotWidth,
  RotHeight         : Integer;
begin
  Theta:=Abs(Angle)*(Pi/180);
  ScaleHeight := Height * ScaleY / 100;
  ScaleWidth := Width * ScaleX / 100;
  RotWidth := Round(Abs(ScaleWidth * Cos(Theta)) - Abs(ScaleHeight * Sin(Theta)));
  RotHeight := Round(Abs(ScaleWidth * Sin(Theta)) + Abs(ScaleHeight * Cos(Theta)));
  Result := Point(RotWidth,RotHeight);
end;
*)

function RelativeAngle(X1,Y1, X2,Y2 : Integer) : Integer;
var
  Theta             : Extended;
  XDist,
  YDist             : Integer;
begin
  Result := 0;

  //arctan((y2-y1)/(x2-x1))
  XDist := X2 - X1;
  YDist := Y1 - Y2;
  if (XDist = 0) and (YDist=0) then exit;

  if YDist=0 then
    Theta := arctan((X2-X1))
  else
    Theta := arctan((X2-X1)/(Y1-Y2));

  Result := Round(RadToDeg(Theta));
  if (X2 >= X1) and (Y2 >= Y1) then //Quadrant = 2
    Result := 90+(90-Abs(Result))
  else
  if (X2 <= X1) and (Y2 >= Y1) then //Quadrant = 3
    Result := 180 + Abs(Result)
  else
  if (X2 <= X1) and (Y2 <= Y1) then //Quadrant = 4
    Result := 270+(90-Abs(Result));
end;


{ TDIBFilter }

constructor TDIBFilter.Create;
begin
  inherited;
  Opacity := 255;
  RedBias := 0;
  GreenBias := 0;
  BlueBias := 0;
  Factor := 0;
end;

class function TDIBFilter.GetDisplayName: String;
begin
  Result := '(Unknown)';
end;

{ TAbstractSuperDIB }

procedure TAbstractSuperDIB.ApplyFilter(AFilter: TDIBFilter);
var
  Dest                        : TWinDIB;
  RedAv,
  GreenAv,
  BlueAv                      : SmallInt;
  AvModulo,
  MatrixOffset,
  LineSize,
  NoLines,
  PixelsPerLine               : DWord;
  Factor                      : Integer;
  MData,
  SourceData,
  DestData                    : Pointer;
  NewData                     : array[0..8] of SmallInt;
  RB,
  GB,
  BB                          : Smallint;
begin
  Dest := TWinDIB.Create(Width,Height);
  LineSize := Width * 4;
  NoLines := Height - 2;
  PixelsPerLine := Width - 2;
  AvModulo := LineSize-12;

  with AFilter do begin
    //Row 1
    NewData[0] := Data[6];
    NewData[1] := Data[7];
    NewData[2] := Data[8];
    //Row 2
    NewData[3] := Data[3];
    NewData[4] := Data[4];
    NewData[5] := Data[5];
    //Row 3
    NewData[6] := Data[0];
    NewData[7] := Data[1];
    NewData[8] := Data[2];
  end;
  if AFilter.Factor > 0 then
    Factor := AFilter.Factor
  else begin
    Factor :=
      NewData[0] + NewData[1] + NewData[2] +
      NewData[3] + NewData[4] + NewData[5] +
      NewData[6] + NewData[7] + NewData[8];
  end;

  MData := @NewData[0];


  MatrixOffset := LineSize + 4;
  SourceData := Pointer( Integer(FData) + MatrixOffset );
  DestData := Pointer( Integer(Dest.FData) + MatrixOffset );

  //Bias
  RB := AFilter.RedBias;
  GB := AFilter.GreenBias;
  BB := AFilter.BlueBias;
  asm
    push  ESI
    push  EDI
    push  EBX

    mov   ESI, SourceData
    mov   EDI, DestData
    mov   ECX, NoLines
  @VLoop:
    push  ECX
    mov   ECX, PixelsPerLine
  @HLoop:
    call  @CalcPixels

    //Mask
    mov   al, [ESI+3]
    shl   EAX, 8
    //Red
    mov   BX, RedAv
    lea   EAX, [EAX+EBX]
    shl   EAX, 8
    //Green
    mov   BX, GreenAv
    lea   EAX, [EAX+EBX]
    shl   EAX,8
    //Blue
    mov   BX, BlueAv
    lea   EAX, [EAX+EBX]

    //Store the pixel color
    mov   [EDI], EAX
    lea   EDI, [EDI+4]
    lea   ESI, [ESI+4]



    dec   ECX
    jnz   @HLoop

    lea   EDI, [EDI+8]
    lea   ESI, [ESI+8]

    pop   ECX
    dec   ECX
    jnz   @VLoop

  @TheEnd:
    pop   EBX
    pop   EDI
    pop   ESI
    jmp   @Exit

    //==================== CALC PIXELS =========================//
  @CalcPixels:
    push  ESI
    push  EDI

    sub   ESI, MatrixOffset
    mov   EDI, MData

  @AvVLoop:
    //==================== LINE 1 =========================//
    //Pixel1
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    mov   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    mov   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    mov   RedAV, AX

    lea   EDI,[EDI+2]

    //Pixel2
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    //Pixel3
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    add   ESI, AvModulo

    //==================== LINE 2 =========================//
    //Pixel1

    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    //Pixel2
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    //Pixel3
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    add   ESI, AvModulo

    //==================== LINE 3 =========================//
    //Pixel1
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX

    lea   EDI,[EDI+2]

    //Pixel2
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    lea   ESI, [ESI+2]
    imul  bx
    Add   RedAv, AX


    lea   EDI,[EDI+2]

    //Pixel3
    mov   bx,  [EDI] //Get filter value

    //BLUE
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   BlueAv, AX

    //GREEN
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    inc   ESI
    imul  bx
    Add   GreenAv, AX

    //RED
    xor   EAX, EAX
    mov   al,  [ESI] //Get pixel values
    imul  bx
    Add   RedAv, AX

    //Now add the RGBM Bias
//    xor  EAX, EAX
    mov  AX, BB
    Add  BlueAv, AX
    mov  AX, GB
    Add  GreenAv, AX
    mov  AX, RB
    Add  RedAv, AX


    //Now divide by the factor
    mov   EBX, Factor
    cmp   EBX, 0
    jz    @NoDivide

    //BLUE
    mov   AX, BlueAv
    cwd
    idiv  BX
    mov   BlueAv, AX

    //GREEN
    mov   AX, GreenAv
    cwd
    idiv  BX
    mov   GreenAv, AX

    //RED
    mov   AX, RedAv
    cwd
    idiv  BX
    mov   RedAv, AX

  @NoDivide:

    //Put RED into 0..255 range
    cmp   RedAv, 255
    jl    @RedLT255
    mov   RedAv, 255
  @RedLT255:
    cmp   RedAv, 0
    jg    @RedGT0
    mov   RedAv, 0
  @RedGt0:

    //Put Green into 0..255 range
    cmp   GreenAv, 255
    jl    @GreenLT255
    mov   GreenAv, 255
  @GreenLT255:
    cmp   GreenAv, 0
    jg    @GreenGT0
    mov   GreenAv, 0
  @GreenGt0:

    //Put Blue into 0..255 range
    cmp   BlueAv, 255
    jl    @BlueLT255
    mov   BlueAv, 255
  @BlueLT255:
    cmp   BlueAv, 0
    jg    @BlueGT0
    mov   BlueAv, 0
  @BlueGt0:

    pop   EDI
    pop   ESI

    ret

  @Exit:

  end;

  Dest.Opacity := AFilter.Opacity;
  Dest.Draw(0,0,Width,Height,Self,0,0);
  Dest.Free;
end;

procedure TAbstractSuperDIB.AssignHeaderTo(Dest: TPersistent);
begin
  if Dest is TAbstractSuperDIB then begin
    TAbstractSuperDIB(Dest).FAngle := Self.FAngle;
    TAbstractSuperDIB(Dest).FAutoSize := Self.FAutoSize;
    TAbstractSuperDIB(Dest).FMasked := Self.FMasked;
    TAbstractSuperDIB(Dest).FOpacity := Self.FOpacity;
    TAbstractSuperDIB(Dest).FScaleX := Self.FScaleX;
    TAbstractSuperDIB(Dest).FScaleY := Self.FScaleY;
    TAbstractSuperDIB(Dest).FTransparent := Self.FTransparent;
    TAbstractSuperDIB(Dest).FTransparentColor := Self.FTransparentColor;
    TAbstractSuperDIB(Dest).FTransparentMode := Self.FTransparentMode;
    TAbstractSuperDIB(Dest).FBlitter := Self.FBlitter;
    if not TAbstractSuperDIB(Dest).FOwnedData then begin
      TAbstractSuperDIB(Dest).FWidth := Self.Width;
      TAbstractSuperDIB(Dest).FHeight := Self.Height;
    end;
  end else
    raise EDIBError.Create('Cannot assign a TAbstractDIB to a ' + Dest.ClassName);
end;

procedure TAbstractSuperDIB.AssignTo(Dest: TPersistent);
begin
  if Dest is TAbstractSuperDIB then begin
    TAbstractSuperDIB(Dest).CopyPicture(Self);
    AssignHeaderTo(Dest);
    TAbstractSuperDIB(Dest).Changed;
  end else
    inherited;
end;

procedure TAbstractSuperDIB.ChangeBlitter;
begin
  if Opacity = 255 then begin
    if Masked then
      FBlitter := MaskedBlit
    else
    if Transparent then
      FBlitter := TransparentBlit
    else
      FBlitter := SolidBlit;
  end else
  begin
    if Masked then
      FBlitter := MaskedBlitO
    else
    if Transparent then
      FBlitter := TransparentBlitO
    else
      FBlitter := SolidBlitO;
  end;
end;

procedure TAbstractSuperDIB.Changed;
begin
  ChangeBlitter;
  if Assigned(FOnChange) then FOnChange(Self);
end;

constructor TAbstractSuperDIB.Create;
begin
  inherited;
  FData := Nil;
  FBlitter := SolidBlit;
  FOwnedData := True;
  ResetHeader;
end;

procedure TAbstractSuperDIB.CopyPicture(Source: TAbstractSuperDIB);
begin
  Resize(Source.Width,Source.Height);
  if (Source.Width > 0) and (Source.Height > 0) then
    Move(Source.FData^,FData^,Width * Height * 4);
end;

constructor TAbstractSuperDIB.Create(aWidth, aHeight: Word);
begin
  Create;
  Resize(aWidth,aHeight);
  ClipRect := Rect(0,0,aWidth,aHeight);
end;

constructor TAbstractSuperDIB.CreateReplicaOf(aSource: TAbstractSuperDIB);
begin
  inherited Create;
  PointDataAt(aSource);
end;

procedure TAbstractSuperDIB.DefineProperties(Filer: TFiler);
begin
  inherited;
  Filer.DefineBinaryProperty('Data',LoadDataFromStream, SaveDataToStream, FData <> nil);
end;



procedure TAbstractSuperDIB.Draw(DestX, DestY, DestWidth,
  DestHeight: Integer; Dest: TAbstractSuperDIB; SrcX, SrcY: Word);
var
  FirstPixel,
  LastPixel,
  FirstLine,
  LastLine          : Integer;
  NoPixels,
  NoLines           : Integer;
  SourceData,
  DestData          : Pointer;
  SourceModulo,
  DestModulo        : DWord;

  Result            : TWinDib;

  Theta             : Extended;
  RotWidth,
  RotHeight         : Integer;
  ScaleWidth,
  ScaleHeight       : Real;

begin
  //if drawn at an angle, we need to handle this first.
  if (Angle <>0) or (ScaleX <> 100) or (ScaleY <> 100) then begin
    if FAutoSize then begin
      Theta:=Abs(Angle)*(Pi/180);
      ScaleHeight := DestHeight * ScaleY / 100;
      ScaleWidth := DestWidth * ScaleX / 100;
      RotWidth := Round(Abs(ScaleWidth * Cos(Theta)) + Abs(ScaleHeight * Sin(Theta)));
      RotHeight := Round(Abs(ScaleWidth * Sin(Theta)) + Abs(ScaleHeight * Cos(Theta)));
      Result := TWinDib.Create(RotWidth,RotHeight);
    end else begin
      Result := TWinDib.Create(DestWidth,DestHeight);
    end;

    try
      RotoZoom(Result);

      Result.Masked := Masked;
      Result.Opacity := Opacity;
      if Transparent then begin
        Result.Transparent := True;
        Result.TransparentMode := tmFixed;
        Result.TransparentColor := TransparentColor;
      end;
      Result.FBlitter := FBlitter;
      Result.Draw(DestX,DestY,Result.Width,Result.Height,Dest,0,0)
    finally
      Result.Free;
    end;
    exit;
  end;

  SrcX := Abs(Srcx);

  if DestX < Dest.ClipRect.Left then begin
    DestWidth := DestWidth - (Dest.ClipRect.Left - DestX);
    SrcX := SrcX + (Dest.ClipRect.Left - DestX);
    DestX := Dest.ClipRect.Left;
  end;

  if DestY < Dest.ClipRect.Top then begin
    DestHeight := DestHeight - (Dest.ClipRect.Top - DestY);
    SrcY := SrcY + (Dest.ClipRect.Top - DestY);
    DestY := Dest.ClipRect.Top;
  end;

  if DestX + DestWidth > Dest.ClipRect.Right then begin
    Dec(DestWidth, (DestX + DestWidth) - Dest.ClipRect.Right -1);
  end;

  if DestY + DestHeight > Dest.ClipRect.Bottom then begin
    Dec(DestHeight, (DestY + DestHeight) - Dest.ClipRect.Bottom -1);
  end;

  if DestWidth + SrcX > Width then DestWidth := Width - SrcX;
  if DestHeight + SrcY > Height then DestHeight := Height - SrcY;
  if DestHeight <= 0 then exit;
  if DestWidth <= 0 then exit;

  if DestX > Dest.Width then exit;
  if DestY > Dest.Height then exit;
  if DestX + Width < 0 then exit;
  if DestY + Height < 0 then exit;

  if DestWidth + DestX > Dest.Width then
    DestWidth := Dest.Width - DestX;
  if DestHeight + DestY > Dest.Height then
    DestHeight := Dest.Height - DestY;

  //FirstPixel
  FirstPixel := SrcX;
  If DestX < 0 then begin
    FirstPixel := FirstPixel + Abs(DestX);
    Dec(DestWidth, FirstPixel);
    if DestWidth < 1 then exit;
    DestX := 0;
  end;

  //LastPixel
  LastPixel := FirstPixel + DestWidth;
  if LastPixel > Width then LastPixel := Width;

  //No of pixels per line
  NoPixels := LastPixel - FirstPixel;
  if NoPixels <1 then exit;

  //First line
  FirstLine := SrcY;
  if DestY < 0 then begin
    FirstLine := FirstLine + Abs(DestY);
    Dec(DestHeight, FirstLine);
    if Destheight < 1 then exit;
    DestY := 0;
  end;



  //Last line
  LastLine := FirstLine + DestHeight;
  if LastLine > Height then LastLine := Height;

  //No of lines
  NoLines := LastLine - FirstLine;
  if NoLines <1 then exit;

  //DIBS are upside down !
  FirstLine := (Height-1) - FirstLine;
  DestY := (Dest.Height-1) - DestY;

  //Work out memory addresses of the first pixel, in source and dest
  SourceData := Pointer( Integer(FData) + (FirstLine * Width *4) + (FirstPixel*4));
  DestData := Pointer( Integer(Dest.FData) + (DestY * Dest.Width *4) + (DestX*4) );

  //Work out the modulos
  SourceModulo := (NoPixels*4) + (Width *4);
  DestModulo := (NoPixels*4) + (Dest.Width *4);

  FBlitter(SourceData, DestData, SourceModulo, DestModulo, NoPixels, NoLines);
  Dest.Changed;
end;

procedure TAbstractSuperDIB.DrawMask(DestX, DestY, DestWidth,
  DestHeight: Integer; Dest: TAbstractSuperDIB; SrcX, SrcY: Word);
var
  OrigBlitter                 : TBlitterProc;
begin
  OrigBlitter := FBlitter;
  try
    FBlitter := BlitMaskOnly;
    Draw(DestX,DestY,DestWidth,DestHeight,Dest,SrcX,SrcY);
  finally
    FBlitter := OrigBlitter;
  end;
end;

function TAbstractSuperDIB.GetTransparentColor: TColor;
begin
  if TransparentMode = tmFixed then
    Result := FTransparentColor
  else
    Result := TColor(FData^);
end;

procedure TAbstractSuperDIB.ImportMask(aFilename: String);
var
  PixelCount        : Cardinal;
  Source1,
  Source2           : TWinDIB;
  SrcData,
  DestData          : Pointer;
begin
  if (Width = 0) or (Height = 0) then
    raise EDIBError.Create('You cannot import a mask until you have an image.');

  Source1 := TWinDib.Create;
  Source2 := TWinDib.Create(Width,Height);
  SrcData := Source2.FData;
  DestData := FData;
  PixelCount := Width * Height;
  try
    Source1.ImportPicture(aFilename);
    StretchBlt(Source2.Handle,0,0,Width,Height,Source1.Handle,0,0,Source1.Width,Source1.Height,SrcCopy);
    asm
          push EDI
          push ESI

          mov  ESI, SrcData
          mov  EDI, DestData
          mov  ECX, PixelCount

    @Loop:
          Mov  al, [ESI]
          mov  [EDI+3], al
          lea  ESI, [ESI+4]
          lea  EDI, [EDI+4]
          dec  ECX
          jnz  @Loop

          pop  ESI
          pop  EDI
    end;
  finally
    Source1.Free;
    Source2.Free;
  end;
  Changed;
end;

procedure TAbstractSuperDIB.ImportPicture(aFilename: String);
var
  Pic               : TPicture;
  WinDIB            : TWinDIB;
begin
  Pic := TPicture.Create;
  WinDIB := TWinDib.Create;
  try
    Pic.LoadFromFile(aFilename);
    Resize(Pic.Width,Pic.Height);
    WinDIB.ReSize(Width,Height);
    WinDib.Canvas.Draw(0,0,Pic.Graphic);
    WinDib.Draw(0,0,Width,Height,Self,0,0);
  finally
    WinDIB.Free;
    Pic.Free;
  end;
  Changed;
end;

procedure TAbstractSuperDIB.LoadDataFromStream(S: TStream);
var
  W,
  H                 : Word;
begin
  with S do begin
    Read(W, SizeOf(W));
    Read(H, SizeOf(H));
    Resize(W,H);
    Read(FData^, Width * Height * 4);
  end;
end;

procedure TAbstractSuperDIB.MaskedBlit(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
begin
  asm
          push EDI
          push ESI
          push EBX

          mov  ESI, SourceData
          mov  EDI, DestData
          xor  EDX, EDX

  @VLoop:
          mov  ECX, NoPixels
  @HLoop:
          mov  bh, [ESI+3]
          mov  bl, [ESI+3]
          not bh

          //Green
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX

          Xor  Ah, Ah
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Blue
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX

          Xor  Ah, Ah
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Red
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX

          Xor  Ah, Ah
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          Inc  ESI
          Inc  EDI

          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  NoLines
          jnz  @VLoop
  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.MaskedBlitO(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
var
  Opacity           : Byte;
begin
  Opacity := FOpacity;
  asm
          push EDI
          push ESI
          push EBX

          mov  ESI, SourceData
          mov  EDI, DestData

          xor  EDX, EDX
  @VLoop:
          mov  ECX, NoPixels
  @HLoop:
          mov  bl, Opacity
          mov  al, [ESI+3]
          mul  bl
          mov  bh, ah
          mov  bl, ah
          not  bh

          //Green
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Blue
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Red
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          Inc  ESI
          inc  EDI

          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  NoLines
          jnz  @VLoop
  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;
end;


procedure TAbstractSuperDIB.QuickFill(aColor: TColor);
var
  NumOfColors       : DWord;
  Area              : Pointer;
begin
  Area := FData;
  aColor := ColorToRGB(aColor);
  NumOfColors := Width * Height;
  asm
    push    EDI
    push    ESI
    push    EBX

    mov     eax, aColor
    bswap   eax
    ror     Eax, 8
    mov     edi, Area
    mov     ecx, NumOfColors
    cld
    rep     StoSD

    pop     EBX
    pop     ESI
    pop     EDI
  end;
  Changed;
end;

procedure TAbstractSuperDIB.QuickFillRect(aColor: TColor; aLeft, aTop,
  aWidth, aHeight: Integer);
var
  NoLines,
  NoPixels,
  DestModulo                  : DWord;
  Data                        : Pointer;
begin
  if aWidth < 1 then exit;
  if aHeight < 1 then exit;
  if aLeft + aWidth < 0 then exit;
  if aTop + aHeight < 0 then exit;
  if aLeft >= Width then exit;
  if aTop >=Height then exit;

  if aLeft < 0 then begin
    aWidth := aWidth - abs(aLeft);
    aLeft := 0;
  end;

  if aTop < 0 then begin
    aHeight := aHeight - abs(aTop);
    aTop := 0;
  end;

  if aLeft + aWidth > Width then aWidth := Width - aLeft;
  if aTop + aHeight > Height then aHeight := Height - aTop;

  NoPixels := aWidth;
  NoLines := aHeight;
  aColor := ColorToRGB(aColor);

  Data := Pointer( Integer(FData) + (((Height-1) - aTop) * Width *4) + (aLeft*4) );
  DestModulo := (NoPixels*4) + (Width *4);

  asm
          push EDI
          push ESI

          mov  EDI, Data
          mov  EDX, NoLines
          mov  EAX, aColor
          bswap EAX
          ror  EAX, 8
  @VLoop:
          mov  ECX, NoPixels
  rep     STOSD

          sub  EDI, DestModulo
          dec  EDX
          jnz  @VLoop
  @TheEnd:
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.ReSize(aWidth, aHeight: Word);
var
  FullSize                    : Boolean;
begin
  if (aWidth = Width) and (aHeight = Height) then exit;
  FullSize :=
    (ClipRect.Left = 0) and (ClipRect.Top = 0) and
    (ClipRect.Right = Width-1) and (ClipRect.Bottom = Height-1);

  FreeTheData;
  if aWidth < 1 then aWidth := 1;
  if aHeight < 1 then aHeight := 1;
  FWidth := aWidth;
  FHeight := aHeight;
  CreateTheData;
//  QuickFill(clBlack);
  if FullSize then
    FClipRect := Rect(0,0,Width-1,Height-1)
  else
    ClipRect := ClipRect;
end;

procedure TAbstractSuperDIB.RotoZoom(D: TAbstractSuperDIB);
var
  Source,
  Dest              : PChar;
  DDX,
  DDY,
  D2X,
  D2Y               : Integer;
  I,
  J                 : Integer;
  Xpos,
  YPos              : Integer;
  SLineSize         : Integer;
  NegHalfSWidth,
  HalfSWidth,
  NegHalfSHeight,
  HalfSHeight       : Integer;
  DestWidth,
  DestHeight        : Integer;
  ScaleX,
  ScaleY            : Real;
  DestTransCol      : TColor;
begin
//  if D is TAbstractSuperDIB then
//    DestTransCol := TAbstractSuperDIB(D).TransparentColor
//  else
//    DestTransCol := $000000;
  If Masked then
    DestTransCol := $000000
  else
    DestTransCol := TransparentColor;

  Source := FData;
  Dest := D.FData;
  SLineSize := Width *4;

  NegHalfSWidth := -(Width div 2);
  NegHalfSHeight := -(Height div 2);
  HalfSWidth := Width div 2;
  HalfSHeight := Height div 2;


  DestWidth := D.Width;
  DestHeight := D.Height;

  ScaleX := FScaleX / 100;
  ScaleY := FScaleY / 100;

  DDX := Round(((CosTable1[Angle]) * Width) / (Width * ScaleX));
  DDY := Round(((SinTable1[Angle]) * Height) / (Height * ScaleY));

  D2X := Round(((CosTable2[Angle]) * Width) / (Width * ScaleX));
  D2Y := Round(((SinTable2[Angle]) * Height) / (Height * ScaleY));

  I := ((-DDX) * (DestWidth div 2)) - (D2X * (DestHeight div 2));
  J := ((-DDY) * (DestWidth div 2)) - (D2Y * (DestHeight div 2));

  Source := Source + (HalfSHeight * SLineSize) +  (HalfSWidth*4);

  asm
          push EDI
          push ESI
          push EBX

          mov  EDI, Dest
          mov  ESI, Source

          mov  ECX, DestHeight// for VLoop := 0 to D.Height -1 do begin
  @VLoop:
          mov  EAX, I
          mov  EBX, J
          mov  XPos, EAX      // XPos := I
          mov  YPos, EBX      // YPos := J

          push ECX
          mov  ECX, DestWidth //for HLoop:=0 to D.Width -1 do begin

  @HLoop:
          //Calculate offset using int(YPos)
          Mov  EAX, YPos
          test EAX, EAX
          jns  @ActualYGTZero
          lea  EAX, [EAX+$000000ff]

  @ActualYGTZero:
          sar  EAX, $08       // ActualY := YPos div 256;

          cmp  EAX, HalfSHeight
          Jge  @SkipPixel     // If ActualY > (S.Height div 2) then skip

          mov  EBX, NegHalfSHeight
//          neg  EBX
          cmp  EAX, EBX
          jle  @SkipPixel     // If ActualY < -(S.Height div 2) then skip

          mov  EBX, SLineSize
          imul EBX
          mov  EDX, EAX       // Offset := (ActualY * SLineSize)

          //Calculate offset using ActualX
          mov  EAX, XPos
          test EAX, EAX
          jns  @ActualXGTZero
          lea  EAX, [EAX+$000000ff]

  @ActualXGTZero:
          sar  EAX, $08       // ActualX := XPos div 256;

          cmp  EAX, HalfSWidth
          Jge  @SkipPixel     //if ActualX > (S.Width div 2) then skip

          mov  EBX, NegHalfSWidth
//          neg  EBX
          cmp  EAX, EBX
          Jle  @SkipPixel

          lea  EAX, [EAX*4+EDX]  //EAX := EAX * 4; {4 bytes per pixel}

          mov  EAX, [ESI +EAX]
          mov  [EDI], EAX

          jmp  @NextPixel

  @SkipPixel:
          mov  EAX, DestTransCol
          mov  [EDI], EAX

  @NextPixel:
          lea  EDI, [EDI+4]
          mov  EAX, ddx
          mov  EBX, ddy
          add  XPos, EAX      //XPos := XPos + DDX;
          add  YPos, EBX      //Ypos := YPos + DDY;
          dec  ECX
          jnz  @HLoop         //end; //for HLoop

  @NextLine:
          pop  ECX
          mov  EAX, D2X
          mov  EBX, D2Y
          add  I, EAX         //I := I + D2X;
          add  J, EBX         //J := J + D2Y;

          dec  ECX
          jnz  @Vloop         //end; //for VLoop

  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;
  D.Changed;
end;

procedure TAbstractSuperDIB.SaveDataToStream(S: TStream);
begin
  with S do begin
    Write(FWidth,SizeOf(FWidth));
    Write(FHeight, SizeOf(FHeight));
    Write(FData^, Width * Height * 4);
  end;
end;

procedure TAbstractSuperDIB.SetAngle(const Value: TAngle);
begin
  if Value = FAngle then exit;
  FAngle := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetAutoSize(const Value: Boolean);
begin
  if Value = FAutosize then exit;
  FAutoSize := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetHeight(const aValue: Word);
begin
  Resize(Width, aValue);
end;

procedure TAbstractSuperDIB.SetMasked(const Value: Boolean);
begin
  if Value = FMasked then exit;
  FMasked := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetMaskedValues(const Opacity: Byte);
var
  PixelCount        : Cardinal;
  Data              : PByteArray;
begin
  PixelCount := FWidth * FHeight;
  Data := PByteArray(FData);
  asm
          push EDI

          mov  EDI, Data
          mov  al,  Opacity
          mov  ECX, PixelCount
    @Loop:
          mov  [EDI+3], al
          lea  EDI, [EDI+4]
          dec  ECX
          jnz  @Loop

          pop  EDI
  end;
  Changed;
end;

procedure TAbstractSuperDIB.SetOpacity(const Value: Byte);
begin
  if Value = FOpacity then exit;
  FOpacity := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetScale(const Value: Real);
begin
  if (Value = ScaleX) and (Value = ScaleY) then exit;
  FScaleX := Value;
  FScaleY := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetScaleX(const Value: Real);
begin
  if Value = ScaleX then exit;
  FScaleX := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetScaleY(const Value: Real);
begin
  if Value = ScaleY then exit;
  FScaleY := Value;
  Changed;
end;


procedure TAbstractSuperDIB.SetTransparent(const Value: Boolean);
begin
  if Value = FTransparent then exit;
  FTransparent := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetTransparentColor(const Value: TColor);
begin
  if Value = TransparentColor then exit;
  FTransparentColor := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetTransparentMode(
  const Value: TTransparentMode);
begin
  if Value = TransparentMode then exit;
  FTransparentMode := Value;
  Changed;
end;

procedure TAbstractSuperDIB.SetWidth(const aValue: Word);
begin
  ReSize(aValue, Height);
end;

procedure TAbstractSuperDIB.SolidBlit(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
begin
  asm
          push EDI
          push ESI

          mov  ESI, SourceData
          mov  EDI, DestData

          mov  EDX, NoLines
  @VLoop:
          mov  ECX, NoPixels
  rep     MOVSD

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  EDX
          jnz  @VLoop
  @TheEnd:
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.SolidBlitO(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
var
  Opacity           : Byte;
begin
  Opacity := FOpacity;
  asm
          push EDI
          push ESI
          push EBX

          mov  bh,  Opacity
          mov  bl,  Opacity

          mov  ESI, SourceData
          mov  EDI, DestData

          not  bh
          xor  EDX, EDX
  @VLoop:
          mov  ECX, NoPixels

  @HLoop:
          //Green
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Blue
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Red
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          Inc  ESI
          Inc  EDI

          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  NoLines
          jnz  @VLoop
  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.TransparentBlit(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
var
  TransColor        : TColor;
begin
  TransColor := TransparentColor;
//  if TransparentMode = tmFixed then
//     TransColor := FTransparentColor
//  else
//    TransColor := TColor(FData^);
  asm
          push EDI
          push ESI
          push EBX

          mov  ESI, SourceData
          mov  EDI, DestData

          mov  EDX, TransColor
          and  EDX, $00ffffff
  @VLoop:
          mov  ECX, NoPixels
  @HLoop:
          Mov  EAX, [ESI]
          and  EAX, $00ffffff
          cmp  EAX, EDX
          je   @SkipPixel
          mov  [EDI], EAX
  @SkipPixel:
          lea  ESI, [ESI+4]
          lea  EDI, [EDI+4]
          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  NoLines
          jnz  @VLoop
  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.TransparentBlitO(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
var
  TransColor        : TColor;
  Opacity           : Byte;
begin
  Opacity := FOpacity;
  TransColor := TransparentColor;
//  if TransparentMode = tmFixed then
//     TransColor := FTransparentColor
//  else
//    TransColor := TColor(FData^);

  asm
          push EDI
          push ESI
          push EBX

          mov  EDX, TransColor
          and  EDX, $00ffffff
          mov  TransColor, EDX

          mov  bh,  Opacity
          mov  bl,  Opacity

          mov  ESI, SourceData
          mov  EDI, DestData

          xor  EDX, EDX

          not  bh
  @VLoop:
          mov  ECX, NoPixels
  @HLoop:
          mov  EAX, [ESI]
          and  EAX, $00ffffff
          cmp  EAX, TransColor
          je   @SkipPixel


          //Green
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Blue
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          //Red
          xor  Ah,Ah
          LodSB
          Mul  Bl
          Mov  DX, AX
          Xor  AX, AX
          Mov  Al, [EDI]
          mul  Bh
          add  Ax, Dx
          mov  Al, Ah
          StoSB
          Inc  ESI
          Inc  EDI

          Jmp  @NextPixel

  @SkipPixel:
          //Next pixel
          lea  ESI, [ESI+4]
          lea  EDI, [EDI+4]
  @NextPixel:
          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  NoLines
          jnz  @VLoop
  @TheEnd:
          pop  EBX
          pop  ESI
          pop  EDI
  end;

end;


procedure TAbstractSuperDIB.PointDataAt(aSource: TAbstractSuperDIB);
begin
  FreeTheData;
  FOwnedData := False;
  aSource.AssignHeaderTo(Self);
  FData := aSource.FData;
end;

function TAbstractSuperDIB.Valid: Boolean;
begin
  Result := (FData <> nil) or not FOwnedData; 
end;

destructor TAbstractSuperDIB.Destroy;
begin
  FreeTheData;
  inherited;
end;

procedure TAbstractSuperDIB.BlitMaskOnly(SourceData, DestData: Pointer;
  SourceModulo, DestModulo: DWord; NoPixels, NoLines: Integer);
begin
  asm
          push EDI
          push ESI

          mov  ESI, SourceData
          mov  EDI, DestData
          lea  ESI, [ESI+3]
          lea  EDI, [EDI+3]

          mov  EDX, NoLines
  @VLoop:
          mov  ECX, NoPixels
  @HLoop:
          mov  al, [ESI]
          lea  ESI, [ESI+4]
          mov  [EDI], al
          lea  EDI, [EDI+4]
          dec  ECX
          jnz  @HLoop

          sub  ESI, SourceModulo
          sub  EDI, DestModulo
          dec  EDX
          jnz  @VLoop
  @TheEnd:
          pop  ESI
          pop  EDI
  end;
end;

procedure TAbstractSuperDIB.CreateTheData;
begin
  if FOwnedData then CreateData;
  FClipRect := Rect(0,0,Width-1,Height-1);
end;

procedure TAbstractSuperDIB.FreeTheData;
begin
  if FOwnedData and (FData <> nil) then FreeData;
end;

procedure TAbstractSuperDIB.ResetHeader;
begin
  FAngle := 0;
  FMasked := False;
  FOpacity := 255;
  FScaleX := 100;
  FScaleY := 100;
  FTransparent := False;
end;

procedure TAbstractSuperDIB.SetClipRect(const aRect: TRect);
var
  Temp                        : Integer;
begin
  FClipRect := aRect;
  with FClipRect do begin
    if Left > Right then begin
      Temp := Left;
      Left := Right;
      Right := Temp;
    end;

    if Top > Bottom then begin
      Temp := Top;
      Top := Bottom;
      Bottom := Temp;
    end;

    if Right >= Width then Right := Width-1;
    if Left < 0 then Left := 0;
    if Bottom >= Height then Bottom := Height-1;
    if Top < 0 then Top := 0;
  end;
end;

procedure TAbstractSuperDIB.StretchCopyPicture(Source: TAbstractSuperDIB);
var
  SWidth, SHeight,
  DWidth, DHeight : DWord;
  XInt, YInt      : DWord;
  XFactor, YFactor: Word;
  SCurrentLine,
  DCurrentLine    : Pointer;
  SLineSize,
  DLineSize       : DWord;

  X,Y             : Word;
  CurrentYFact    : Word;
begin
  SWidth := Source.Width;
  SHeight := Source.Height;
  DWidth := Width;
  DHeight := Height;

  XInt := (SWidth shl 16) div (DWidth shl 16);
  XFactor := (SWidth shl 16) div DWidth;
  YInt := (SHeight shl 16) div (DHeight shl 16);
  YFactor := (SHeight shl 16) div DHeight;


  SCurrentLine := Source.FData;
  DCurrentLine := FData;

  SLineSize := Source.Width * 4;
  DLineSize := Width * 4;
  XInt := XInt * 4;
  YInt := YInt * SLineSize;

  CurrentYFact := 65535;
  asm
        push ESI
        push EDI
        push EBX

        //Current X and Y factor in EDX
        mov  DX, YFactor
        BSWAP EDX
        mov  DX, XFactor

        //X and Y factor in EBX
        mov  BX, YFactor
        BSWAP EBX

        //For Y := 0 to dst.height -1
        mov  ECX, DHeight

  @VLoop:

        //CurrentFactX := 65535;
        mov  BX, 65535
        //SData := SCurrentLine;
        mov  ESI, SCurrentLine
        //DData := DCurrentLine;
        mov  EDI, DCurrentLine

        //For X:=0 to Dst.Width-1
        push ECX
        mov  ECX, DWidth
  @HLoop:
        //DWord(DData^) := DWord(SData^);
        mov  EAX, [ESI]
        mov  [EDI], EAX

        //SData := Pointer(Cardinal(SData) + XInt);
        add  ESI, XInt
        //DData := Pointer(Cardinal(DData) + 4);
        lea  EDI, [EDI+4]

        //Check the XFactor
        sub  BX, DX
        jae  @NoWrapX
        lea  ESI, [ESI+4]
  @NoWrapX:
        dec  ECX
        jnz  @HLoop

        //Switch to Factors for Y loop
        BSWAP EBX
        BSWAP EDX

        //SCurrentLine := Pointer(Cardinal(SCurrentLine) + YInt);
        mov  EAX, YInt
        add  SCurrentLine, EAX
        //DCurrentLine := Pointer(Cardinal(DCurrentLine) + DLineSize);
        mov  EAX, DLineSize
        add  DCurrentLine, EAX

        //Check the Y Factor
        sub  BX, DX
        jae  @NoWrapY
        mov  EAX, SLineSize
        add  SCurrentLine, EAX
  @NoWrapY:
        //Switch back to Factors for X loop
        BSWAP EBX
        BSWAP EDX

        pop  ECX
        dec  ECX
        jnz  @VLoop

  @TheEnd:
        pop  EBX
        pop  EDI
        pop  ESI
  end;
end;

procedure TAbstractSuperDIB.Render8Bit(DestDC: HDC; X, Y, aWidth,
  aHeight: Integer; XSrc, YSrc: Word; ROP: Cardinal; Palette: TDIBPalette);
var
  SourceModulo,
  DestModulo                  : DWord;
  BitInfo                     : PBitmapInfo;
  DestData,
  SourceData                  : Pointer;
  FirstLine,
  LineSize                    : DWord;
  Table                       : Pointer;
  //For temp DIB
  Data                        : Pointer;
  DC                          : HDC;
  OldBitmap,
  Bitmap                      : HBitmap;
begin
  if Palette.UseTable = False then
    raise EDIBError.Create('Render8Bit can only be used with a palette containing a lookup table');
  if GetDeviceCaps(DestDC,BITSPixel) <> 8 then
    raise EDIBError.Create('Render8Bit can only be used to blit to an 8 bit DC');


  if XSrc >= Width then exit;
  if YSrc >= Height then exit;

  if X < 0 then begin
    aWidth := Width - Abs(X);
    X := 0;
  end;
  if Y < 0 then begin
    aHeight := Height - Abs(Y);
    Y := 0;
  end;
  if aWidth < 1 then exit;
  if aHeight < 1 then exit;

  if XSrc + aWidth > Width then aWidth := Width - (aWidth-XSrc);
  if YSrc + aHeight > Height then aHeight := Height - (aHeight-YSrc);

  //Make the 8 bit DIB
  DC := CreateCompatibleDC(DestDC);
  GetMem(BitInfo, SizeOf(TBitmapInfo) + 1024);
  with BitInfo.bmiHeader do
  begin
    biSize:=SizeOf(TBitmapInfoHeader);
    biPlanes:=1;
    biBitCount:=8;
    biCompression:=BI_RGB;
    biWidth:=aWidth;
    biHeight:=aHeight;
    biSizeImage:=0;
    biXPelsPerMeter:=0;
    biYPelsPerMeter:=0;
    biClrUsed:=0;
    biClrImportant:=0;
  end;
  GetPaletteEntries(Palette.Palette,0,235,BitInfo.bmiColors[0]);
  SetLastError(0);
  Bitmap := CreateDIBSection(DC,BitInfo^,DIB_PAL_COLORS,Data,0,0);
  if Bitmap = 0 then
    raise Exception.Create('Windows error (' + IntToStr(GetLastError) + ') ' + LastError);
  GDIFlush;
  OldBitmap := SelectObject(DC,Bitmap);

  //We now should have the bits in DATA
  LineSize := aWidth;
  if LineSize mod 4 > 0 then
    LineSize := LineSize + (4-(LineSize mod 4));  //Increase to DWord

  DestModulo := aWidth + LineSize;
  SourceModulo := (aWidth *4) + (Width * 4);

  FirstLine := (Height-1) - YSrc;
  SourceData := Pointer( Integer(FData) + (FirstLine * Width *4) + (XSrc*4));
  DestData := Pointer( Integer(Data) + ( LineSize * (aHeight-1)) );
  Table := @Palette.ColorTable[0];

  asm
      push ESI
      push EDI
      push EBX

      mov  EBX, Table
      mov  ESI, SourceData
      mov  EDI, DestData

      mov  ECX, aHeight
  @YLoop:
      push ECX
      mov  ECX, aWidth
  @XLoop:
      xor  EAX, EAX
      LodSB
      shr  EAX, 2
//      shl  eax, 12
      mov  EDX, EAX

      xor  EAX, EAX
      LodSB //Green
      shr  EAX, 2
      shl  EAX, 6
      add  EDX, EAX

      xor  EAX, EAX
      LodSB //Blue
      shr  EAX, 2
      shl  EAX, 12
      Add  EDX, EAX

      LodSB  //Ignore the mask

      //Get the index number
      mov  al, [EBX+EDX]

      StoSB //Put the result into the 8bit bitmap

      dec  ECX
      Jnz  @XLoop

      Sub  ESI, SourceModulo
      Sub  EDI, DestModulo

      pop  ECX
      dec  ECX
      jnz  @YLoop

  @TheEnd:
      pop  EBX
      pop  EDI
      pop  ESI
  end;

  SetDIBitsToDevice(DestDC, X, Y, aWidth, aHeight, 0,0,0,aHeight,Data,BitInfo^,DIB_RGB_COLORS);
//  bitblt(destdc,x,y,awidth,aheight,dc,0,0,rop);
  SelectObject(DC,OldBitmap);
  DeleteObject(Bitmap);
  DeleteDC(DC);
  FreeMem(BitInfo);
end;

//MakeRGN and MakeRGNFromColor are almost identical.  A bug in one will most
//likely mean that there is a bug in the other.
function TAbstractSuperDIB.MakeRGN(const aMasklevel : Byte) : HRGN;
var
  hData                       : hGlobal;
  pData                       : PRGNData;
  RectCount,
  RectsLeft                   : Integer;
  pPixel,
  pPixel2                     : PByte;
  lCurrentY,
  lWidth,
  X1,
  X2                          : Integer;
  ThisRect                    : PRect;
  TempRGN                     : HRGN;
begin
  Result := 0;

  lWidth := Width-1;
  lCurrentY := Height-1;

  RectCount := cRectAllocs;
  RectsLeft := cRectAllocs;

  hData := GlobalAlloc(GMEM_MOVEABLE, sizeof(TRgnDataHeader) + SizeOf(TRect) * RectCount);
  pData := GlobalLock(hData);

  ThisRect := PRect(@pData^.Buffer[0]);

  pPixel := PByte(Integer(FData)+3);
  X1:=0;
  while lCurrentY >= 0 do begin
    while (pPixel^ <= aMaskLevel) and (X1 < lWidth) do begin
      pPixel := Pointer(DWord(pPixel) + 4);  //Inc (MakeRGNFromColor) adds 16 for some reason
      Inc(X1);
    end;

    X2 := x1;
    pPixel2 := PByte(Integer(pPixel)+4);
    while (X2 < lWidth) and (pPixel2^ > aMaskLevel) do begin
      pPixel := Pointer(DWord(pPixel) + 4);  //Inc (MakeRGNFromColor) adds 16 for some reason
      Inc(X2);
    end;

    if RectsLeft = 0 then begin
      RectsLeft := cRectAllocs;

      if RectCount >= cMaxRectChunks then begin
        with pData^.rdh do begin
          dwSize := SizeOf(TRgnDataHeader);
          iType := RDH_RECTANGLES;
          nCount := 0;
          nRgnSize := 0;
          SetRect(rcBound, 0, 0, lWidth, Height-1);
          nCount := RectCount;
        end;

        TempRGN := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), pData^);
        Assert(TempRGN <> 0);
        if Result <> 0 then begin
          CombineRgn(Result, Result, TempRGN, RGN_OR);
          DeleteObject(TempRGN);
        end else
          Result := TempRGN;

        RectCount := cRectAllocs;
        ThisRect := PRect(@pData^.Buffer[0]);
      end else
        Inc(RectCount, cRectAllocs);

      GlobalUnlock(hData);
      hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), GMEM_MOVEABLE);
      pData := GlobalLock(hData);
      Assert(pData <> NIL);
      ThisRect := PRect(@pData^.Buffer[(RectCount - cRectAllocs) * Sizeof(TRect)]);
    end;

    SetRect(ThisRect^, X1, lCurrentY, X2, lCurrentY+1);
    ThisRect := PRect(Integer(ThisRect) + SizeOf(TRect));
    Dec(RectsLeft);

    pPixel := pPixel2;
    X1 := X2+1;

    if X1 > lWidth then begin
      Dec(lCurrentY);
      X1 := 0;
    end;
  end;

  if RectsLeft > 0 then begin
    Inc(RectCount, cRectAllocs - RectsLeft);
    GlobalUnlock(hData);
    hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), GMEM_MOVEABLE);
    pData := GlobalLock(hData);
    Assert(pData <> NIL);
  end;

  with pData^.rdh do begin
    dwSize := SizeOf(TRgnDataHeader);
    iType := RDH_RECTANGLES;
    nCount := 0;
    nRgnSize := 0;
    SetRect(rcBound, 0, 0, lWidth, Height-1);
    nCount := RectCount;
  end;

  TempRGN := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), pData^);
  Assert(TempRGN <> 0);
  if Result <> 0 then begin
    CombineRgn(Result, Result, TempRGN, RGN_OR);
    DeleteObject(TempRGN);
  end else
    Result := TempRGN;

  GlobalUnlock(hData);
  GlobalFree(hData);
end;

//MakeRGN and MakeRGNFromColor are almost identical.  A bug in one will most
//likely mean that there is a bug in the other.
function TAbstractSuperDIB.MakeRGNFromColor(aColor: TColor): HRGN;
var
  hData                       : hGlobal;
  pData                       : PRGNData;
  RectCount,
  RectsLeft                   : Integer;
  pPixel,
  pPixel2                     : PDWord;
  lCurrentY,
  lWidth,
  X1,
  X2                          : Integer;
  ThisRect                    : PRect;
  TempRGN                     : HRGN;
begin
  Result := 0;
  aColor := ColorToRGB(aColor);
  asm
    mov  EAX, aColor
    BSWAP EAX
    mov  aColor, EAX
    shr EAX,8
    mov aColor, EAX
  end;


  lWidth := Width-1;
  lCurrentY := Height-1;

  RectCount := cRectAllocs;
  RectsLeft := cRectAllocs;

  hData := GlobalAlloc(GMEM_MOVEABLE, sizeof(TRgnDataHeader) + SizeOf(TRect) * RectCount);
  pData := GlobalLock(hData);

  ThisRect := PRect(@pData^.Buffer[0]);

  pPixel := FData;
  X1:=0;
  while lCurrentY >= 0 do begin
    while ( (pPixel^ and $00FFFFFF) = aColor) and (X1 < lWidth) do begin
      pPixel := Pointer(DWord(pPixel) + 4);  //Inc adds 16 for some reason
      Inc(X1);
    end;

    X2 := x1;
    pPixel2 := PDWord(Integer(pPixel)+4);
    while (X2 < lWidth) and ( (pPixel2^ and $00FFFFFF) <> aColor) do begin
      pPixel2 := Pointer(DWord(pPixel2) + 4);  //Inc adds 16 for some reason
      Inc(X2);
    end;

    if RectsLeft = 0 then begin
      RectsLeft := cRectAllocs;

      if RectCount >= cMaxRectChunks then begin
        with pData^.rdh do begin
          dwSize := SizeOf(TRgnDataHeader);
          iType := RDH_RECTANGLES;
          nCount := 0;
          nRgnSize := 0;
          SetRect(rcBound, 0, 0, lWidth, Height-1);
          nCount := RectCount;
        end;

        TempRGN := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), pData^);
        Assert(TempRGN <> 0);
        if Result <> 0 then begin
          CombineRgn(Result, Result, TempRGN, RGN_OR);
          DeleteObject(TempRGN);
        end else
          Result := TempRGN;

        RectCount := cRectAllocs;
        ThisRect := PRect(@pData^.Buffer[0]);
      end else
        Inc(RectCount, cRectAllocs);

      GlobalUnlock(hData);
      hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), GMEM_MOVEABLE);
      pData := GlobalLock(hData);
      Assert(pData <> NIL);
      ThisRect := PRect(@pData^.Buffer[(RectCount - cRectAllocs) * Sizeof(TRect)]);
    end;

    SetRect(ThisRect^, X1, lCurrentY, X2, lCurrentY+1);
    ThisRect := PRect(Integer(ThisRect) + SizeOf(TRect));
    Dec(RectsLeft);

    pPixel := pPixel2;
    X1 := X2+1;

    if X1 > lWidth then begin
      Dec(lCurrentY);
      X1 := 0;
    end;
  end;

  if RectsLeft > 0 then begin
    Inc(RectCount, cRectAllocs - RectsLeft);
    GlobalUnlock(hData);
    hData := GlobalReAlloc(hData, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), GMEM_MOVEABLE);
    pData := GlobalLock(hData);
    Assert(pData <> NIL);
  end;

  with pData^.rdh do begin
    dwSize := SizeOf(TRgnDataHeader);
    iType := RDH_RECTANGLES;
    nCount := 0;
    nRgnSize := 0;
    SetRect(rcBound, 0, 0, lWidth, Height-1);
    nCount := RectCount;
  end;

  TempRGN := ExtCreateRegion(NIL, SizeOf(TRgnDataHeader) + (RectCount * SizeOf(TRect)), pData^);
  Assert(TempRGN <> 0);
  if Result <> 0 then begin
    CombineRgn(Result, Result, TempRGN, RGN_OR);
    DeleteObject(TempRGN);
  end else
    Result := TempRGN;

  GlobalUnlock(hData);
  GlobalFree(hData);
end;


{ TCustomWinDIB }

constructor TCustomWinDIB.Create;
begin
  inherited;
  FCanvas := TCanvas.Create;
end;

procedure TCustomWinDIB.CreateData;
var
  TempDC            : HDC;
  BitInfo           : TBitmapInfo;
begin
  TempDC := GetDC(0);
  FDC := CreateCompatibleDC(TempDC);
  with BitInfo.bmiHeader do
  begin
    biSize:=SizeOf(TBitmapInfoHeader);
    biPlanes:=1;
    biBitCount:=32;
    biCompression:=BI_RGB;
    biWidth:=Width;
    biHeight:=Height;
    biSizeImage:=0;
    biXPelsPerMeter:=0;
    biYPelsPerMeter:=0;
    biClrUsed:=0;
    biClrImportant:=0;
  end;
  SetLastError(0);
  FBitmap := CreateDIBSection(FDC,BitInfo,DIB_RGB_COLORS,FData,0,0);
  if FBitmap = 0 then
    raise Exception.Create('Windows error (' + IntToStr(GetLastError) + ') ' + LastError);
  GDIFlush;

  //SysRPL
  //DeleteObject(SelectObject(FDC,FBitmap));
  FOldBitmap := SelectObject(FDC,FBitmap);
  ReleaseDC(0,TempDC);
  FCanvas.Handle := FDC;
end;

destructor TCustomWinDIB.Destroy;
begin
  FCanvas.Free;
  FCanvas := nil;
  inherited;
end;

procedure TCustomWinDIB.FreeData;
begin
  GDIFlush;
  SelectObject(FDC,FOldBitmap);
  DeleteObject(FBitmap);
  DeleteDC(FDC);
  //SysRPL
  FData := nil;
end;


{ TCustomMemoryDIB }

procedure TCustomMemoryDIB.CreateData;
begin
  Getmem(FData,Width * Height *4);
end;

procedure TCustomMemoryDIB.FreeData;
begin
  Freemem(FData);
end;


{==================INITIALIZATION====================}
var
  X                 : TAngle;
  Angle             : Real;

initialization
  X:=0;
  for X:=lo(X) to High(X) do begin
    Angle := X * pi/((High(X)+1) div 2);
    SinTable1[X] := Round(Sin(Angle)*256);
    CosTable1[X] := Round(Cos(Angle)*256);
    SinTable2[X] := Round(Sin(Angle + (pi/2))*256);
    CosTable2[X] := Round(Cos(Angle + (pi/2))*256);
 end;
end.
